@@ -1,5 +1,3 @@
-from typing import List
-
 from snowflake.ml.feature_store import Entity
 
 zipcode_entity = Entity(
@@ -12,5 +10,5 @@ plane_entity = Entity(name="PLANE_MODEL", join_keys=["PLANE_MODEL"], desc="The m
 
 
 # This will be invoked by example_helper.py. Do not change function name.
-def get_all_entities() -> List[Entity]:
+def get_all_entities() -> list[Entity]:
     return [zipcode_entity, plane_entity]
