@@ -1,12 +1,13 @@
 from __future__ import annotations
 
 import json
+import logging
 import re
 import warnings
 from collections import OrderedDict
 from dataclasses import asdict, dataclass
 from enum import Enum
-from typing import Any, Dict, List, Optional, Union
+from typing import Any, Optional, Union
 
 from snowflake.ml._internal.exceptions import (
     error_codes,
@@ -31,10 +32,12 @@ from snowflake.snowpark.types import (
     _NumericType,
 )
 
+_DEFAULT_TARGET_LAG = "10 seconds"
 _FEATURE_VIEW_NAME_DELIMITER = "$"
 _LEGACY_TIMESTAMP_COL_PLACEHOLDER_VALS = ["FS_TIMESTAMP_COL_PLACEHOLDER_VAL", "NULL"]
 _TIMESTAMP_COL_PLACEHOLDER = "NULL"
 _FEATURE_OBJ_TYPE = "FEATURE_OBJ_TYPE"
+_ONLINE_TABLE_SUFFIX = "$ONLINE"
 # Feature view version rule is aligned with dataset version rule in SQL.
 _FEATURE_VIEW_VERSION_RE = re.compile(r"^[a-zA-Z0-9][a-zA-Z0-9_.\-]*$")
 _FEATURE_VIEW_VERSION_MAX_LENGTH = 128
@@ -45,11 +48,49 @@ _RESULT_SCAN_QUERY_PATTERN = re.compile(
 )
 
 
+@dataclass(frozen=True)
+class OnlineConfig:
+    """Configuration for online feature storage."""
+
+    enable: bool = False
+    target_lag: Optional[str] = None
+
+    def __post_init__(self) -> None:
+        if self.target_lag is None:
+            return
+        if not isinstance(self.target_lag, str) or not self.target_lag.strip():
+            raise ValueError("target_lag must be a non-empty string")
+
+        object.__setattr__(self, "target_lag", self.target_lag.strip())
+
+    def to_json(self) -> str:
+        data: dict[str, Any] = asdict(self)
+        return json.dumps(data)
+
+    @classmethod
+    def from_json(cls, json_str: str) -> OnlineConfig:
+        data = json.loads(json_str)
+        return cls(**data)
+
+
+class StoreType(Enum):
+    """
+    Enumeration for specifying the storage type when reading from or refreshing feature views.
+
+    The Feature View supports two storage modes:
+    - OFFLINE: Traditional batch storage for historical feature data and training
+    - ONLINE: Low-latency storage optimized for real-time feature serving
+    """
+
+    ONLINE = "online"
+    OFFLINE = "offline"
+
+
 @dataclass(frozen=True)
 class _FeatureViewMetadata:
     """Represent metadata tracked on top of FV backend object"""
 
-    entities: List[str]
+    entities: list[str]
     timestamp_col: str
 
     def to_json(self) -> str:
@@ -73,7 +114,7 @@ class _CompactRepresentation:
     sch: str
     name: str
     version: str
-    feature_indices: Optional[List[int]] = None
+    feature_indices: Optional[list[int]] = None
 
     def to_json(self) -> str:
         return json.dumps(asdict(self))
@@ -110,7 +151,7 @@ class FeatureViewStatus(Enum):
 @dataclass(frozen=True)
 class FeatureViewSlice:
     feature_view_ref: FeatureView
-    names: List[SqlIdentifier]
+    names: list[SqlIdentifier]
 
     def __repr__(self) -> str:
         states = (f"{k}={v}" for k, v in vars(self).items())
@@ -148,7 +189,7 @@ class FeatureViewSlice:
             feature_indices=self._feature_names_to_indices(),
         )
 
-    def _feature_names_to_indices(self) -> List[int]:
+    def _feature_names_to_indices(self) -> list[int]:
         name_to_indices_map = {name: idx for idx, name in enumerate(self.feature_view_ref.feature_names)}
         return [name_to_indices_map[n] for n in self.names]
 
@@ -161,7 +202,7 @@ class FeatureView(lineage_node.LineageNode):
     def __init__(
         self,
         name: str,
-        entities: List[Entity],
+        entities: list[Entity],
         feature_df: DataFrame,
         *,
         timestamp_col: Optional[str] = None,
@@ -170,7 +211,8 @@ class FeatureView(lineage_node.LineageNode):
         warehouse: Optional[str] = None,
         initialize: str = "ON_CREATE",
         refresh_mode: str = "AUTO",
-        cluster_by: Optional[List[str]] = None,
+        cluster_by: Optional[list[str]] = None,
+        online_config: Optional[OnlineConfig] = None,
         **_kwargs: Any,
     ) -> None:
         """
@@ -198,18 +240,20 @@ class FeatureView(lineage_node.LineageNode):
                 after you register the feature view. It supports ON_CREATE (default) or ON_SCHEDULE. ON_CREATE refreshes
                 the feature view synchronously at creation. ON_SCHEDULE refreshes the feature view at the next scheduled
                 refresh. It is only effective when refresh_freq is not None.
-            refresh_mode: The refresh mode of managed feature view. The value can be 'AUTO', 'FULL' or 'INCREMENETAL'.
+            refresh_mode: The refresh mode of managed feature view. The value can be 'AUTO', 'FULL' or 'INCREMENTAL'.
                 For managed feature view, the default value is 'AUTO'. For static feature view it has no effect.
                 Check https://docs.snowflake.com/en/sql-reference/sql/create-dynamic-table for for details.
             cluster_by: Columns to cluster the feature view by.
                 - Defaults to the join keys from entities.
                 - If `timestamp_col` is provided, it is added to the default clustering keys.
+            online_config: Optional configuration for online storage. If provided with enable=True,
+                online storage will be enabled. Defaults to None (no online storage).
             _kwargs: reserved kwargs for system generated args. NOTE: DO NOT USE.
 
         Example::
 
             >>> fs = FeatureStore(...)
-            >>> # draft_fv is a local object that hasn't materiaized to Snowflake backend yet.
+            >>> # draft_fv is a local object that hasn't materialized to Snowflake backend yet.
             >>> feature_df = session.sql("select f_1, f_2 from source_table")
             >>> draft_fv = FeatureView(
             ...     name="my_fv",
@@ -227,12 +271,29 @@ class FeatureView(lineage_node.LineageNode):
             >>> registered_fv = fs.register_feature_view(draft_fv, "v1")
             >>> print(registered_fv.status)
             FeatureViewStatus.ACTIVE
+            <BLANKLINE>
+            >>> # Example with online configuration for online feature storage
+            >>> config = OnlineConfig(enable=True, target_lag='15s')
+            >>> online_fv = FeatureView(
+            ...     name="my_online_fv",
+            ...     entities=[e1, e2],
+            ...     feature_df=feature_df,
+            ...     timestamp_col='TS',
+            ...     refresh_freq='1d',
+            ...     desc='Feature view with online storage',
+            ...     online_config=config  # optional, enables online feature storage
+            ... )
+            >>> registered_online_fv = fs.register_feature_view(online_fv, "v1")
+            >>> print(registered_online_fv.online)
+            True
 
         # noqa: DAR401
         """
+        if online_config is not None:
+            logging.warning("'online_config' is in private preview since 1.12.0. Do not use it in production.")
 
         self._name: SqlIdentifier = SqlIdentifier(name)
-        self._entities: List[Entity] = entities
+        self._entities: list[Entity] = entities
         self._feature_df: DataFrame = feature_df
         self._timestamp_col: Optional[SqlIdentifier] = (
             SqlIdentifier(timestamp_col) if timestamp_col is not None else None
@@ -254,9 +315,10 @@ class FeatureView(lineage_node.LineageNode):
         self._refresh_mode: Optional[str] = refresh_mode
         self._refresh_mode_reason: Optional[str] = None
         self._owner: Optional[str] = None
-        self._cluster_by: List[SqlIdentifier] = (
+        self._cluster_by: list[SqlIdentifier] = (
             [SqlIdentifier(col) for col in cluster_by] if cluster_by is not None else self._get_default_cluster_by()
         )
+        self._online_config: Optional[OnlineConfig] = online_config
 
         # Validate kwargs
         if _kwargs:
@@ -264,7 +326,7 @@ class FeatureView(lineage_node.LineageNode):
 
         self._validate()
 
-    def slice(self, names: List[str]) -> FeatureViewSlice:
+    def slice(self, names: list[str]) -> FeatureViewSlice:
         """
         Select a subset of features within the FeatureView.
 
@@ -343,7 +405,7 @@ class FeatureView(lineage_node.LineageNode):
             raise RuntimeError(f"FeatureView {self.name} has not been registered.")
         return f"{self._database}.{self._schema}.{FeatureView._get_physical_name(self.name, self.version)}"
 
-    def attach_feature_desc(self, descs: Dict[str, str]) -> FeatureView:
+    def attach_feature_desc(self, descs: dict[str, str]) -> FeatureView:
         """
         Associate feature level descriptions to the FeatureView.
 
@@ -396,7 +458,7 @@ class FeatureView(lineage_node.LineageNode):
         return self._name
 
     @property
-    def entities(self) -> List[Entity]:
+    def entities(self) -> list[Entity]:
         return self._entities
 
     @property
@@ -408,7 +470,7 @@ class FeatureView(lineage_node.LineageNode):
         return self._timestamp_col
 
     @property
-    def cluster_by(self) -> Optional[List[SqlIdentifier]]:
+    def cluster_by(self) -> Optional[list[SqlIdentifier]]:
         return self._cluster_by
 
     @property
@@ -463,13 +525,38 @@ class FeatureView(lineage_node.LineageNode):
         return self._status
 
     @property
-    def feature_names(self) -> List[SqlIdentifier]:
+    def feature_names(self) -> list[SqlIdentifier]:
         return list(self._feature_desc.keys()) if self._feature_desc is not None else []
 
     @property
-    def feature_descs(self) -> Optional[Dict[SqlIdentifier, str]]:
+    def feature_descs(self) -> Optional[dict[SqlIdentifier, str]]:
         return self._feature_desc
 
+    @property
+    def online(self) -> bool:
+        return self._online_config.enable if self._online_config else False
+
+    @property
+    def online_config(self) -> Optional[OnlineConfig]:
+        return self._online_config
+
+    def fully_qualified_online_table_name(self) -> str:
+        """Get the fully qualified name for the online feature table.
+
+        Returns:
+            The fully qualified name (<database_name>.<schema_name>.<online_table_name>) for the
+            online feature table in Snowflake.
+
+        Raises:
+            RuntimeError: if the FeatureView is not registered or not configured for online storage.
+        """
+        if self.status == FeatureViewStatus.DRAFT or self.version is None:
+            raise RuntimeError(f"FeatureView {self.name} has not been registered.")
+        if not self.online:
+            raise RuntimeError(f"FeatureView {self.name} is not configured for online storage.")
+        online_table_name = self._get_online_table_name(self.name, self.version)
+        return f"{self._database}.{self._schema}.{online_table_name}"
+
     def list_columns(self) -> DataFrame:
         """List all columns and their information.
 
@@ -687,7 +774,7 @@ Got {len(self._feature_df.queries['queries'])}: {self._feature_df.queries['queri
         if self._initialize not in ["ON_CREATE", "ON_SCHEDULE"]:
             raise ValueError("'initialize' only supports ON_CREATE or ON_SCHEDULE.")
 
-    def _get_column_names(self) -> Optional[List[SqlIdentifier]]:
+    def _get_column_names(self) -> Optional[list[SqlIdentifier]]:
         try:
             return to_sql_identifiers(self._infer_schema_df.columns)
         except SnowparkSQLException as e:
@@ -699,7 +786,7 @@ Got {len(self._feature_df.queries['queries'])}: {self._feature_df.queries['queri
             )
             return None
 
-    def _get_feature_names(self) -> Optional[List[SqlIdentifier]]:
+    def _get_feature_names(self) -> Optional[list[SqlIdentifier]]:
         join_keys = [k for e in self._entities for k in e.join_keys]
         ts_col = [self._timestamp_col] if self._timestamp_col is not None else []
         feature_names = self._get_column_names()
@@ -733,7 +820,7 @@ Got {len(self._feature_df.queries['queries'])}: {self._feature_df.queries['queri
             and self._owner == other._owner
         )
 
-    def _to_dict(self) -> Dict[str, str]:
+    def _to_dict(self) -> dict[str, str]:
         fv_dict = self.__dict__.copy()
         if "_feature_df" in fv_dict:
             fv_dict.pop("_feature_df")
@@ -756,6 +843,8 @@ Got {len(self._feature_df.queries['queries'])}: {self._feature_df.queries['queri
                 feature_desc_dict[k.identifier()] = v
             fv_dict["_feature_desc"] = feature_desc_dict
 
+        fv_dict["_online_config"] = self._online_config.to_json() if self._online_config is not None else None
+
         lineage_node_keys = [key for key in fv_dict if key.startswith("_node") or key == "_session"]
 
         for key in lineage_node_keys:
@@ -844,6 +933,9 @@ Got {len(self._feature_df.queries['queries'])}: {self._feature_df.queries['queri
             owner=json_dict["_owner"],
             infer_schema_df=session.sql(json_dict.get("_infer_schema_query", None)),
             session=session,
+            online_config=OnlineConfig.from_json(json_dict["_online_config"])
+            if json_dict.get("_online_config")
+            else None,
         )
 
     def _get_compact_repr(self) -> _CompactRepresentation:
@@ -898,13 +990,13 @@ Got {len(self._feature_df.queries['queries'])}: {self._feature_df.queries['queri
     @staticmethod
     def _construct_feature_view(
         name: str,
-        entities: List[Entity],
+        entities: list[Entity],
         feature_df: DataFrame,
         timestamp_col: Optional[str],
         desc: str,
         version: str,
         status: FeatureViewStatus,
-        feature_descs: Dict[str, str],
+        feature_descs: dict[str, str],
         refresh_freq: Optional[str],
         database: str,
         schema: str,
@@ -915,7 +1007,8 @@ Got {len(self._feature_df.queries['queries'])}: {self._feature_df.queries['queri
         owner: Optional[str],
         infer_schema_df: Optional[DataFrame],
         session: Session,
-        cluster_by: Optional[List[str]] = None,
+        cluster_by: Optional[list[str]] = None,
+        online_config: Optional[OnlineConfig] = None,
     ) -> FeatureView:
         fv = FeatureView(
             name=name,
@@ -925,6 +1018,7 @@ Got {len(self._feature_df.queries['queries'])}: {self._feature_df.queries['queri
             desc=desc,
             _infer_schema_df=infer_schema_df,
             cluster_by=cluster_by,
+            online_config=online_config,
         )
         fv._version = FeatureViewVersion(version) if version is not None else None
         fv._status = status
@@ -944,7 +1038,7 @@ Got {len(self._feature_df.queries['queries'])}: {self._feature_df.queries['queri
         return fv
 
     #
-    def _get_default_cluster_by(self) -> List[SqlIdentifier]:
+    def _get_default_cluster_by(self) -> list[SqlIdentifier]:
         """
         Get default columns to cluster the feature view by.
         Default cluster_by columns are join keys from entities and timestamp_col if it exists
@@ -961,5 +1055,33 @@ Got {len(self._feature_df.queries['queries'])}: {self._feature_df.queries['queri
 
         return default_cluster_by_cols
 
+    @staticmethod
+    def _get_online_table_name(
+        feature_view_name: Union[SqlIdentifier, str], version: Optional[Union[FeatureViewVersion, str]] = None
+    ) -> SqlIdentifier:
+        """Get the online feature table name without qualification.
+
+        Args:
+            feature_view_name: Offline feature view name.
+            version: Feature view version. If not provided, feature_view_name must be a SqlIdentifier.
+
+        Returns:
+            The online table name SqlIdentifier
+        """
+        if version is None:
+            assert isinstance(feature_view_name, SqlIdentifier), "Single argument must be SqlIdentifier"
+            online_name = f"{feature_view_name.resolved()}{_ONLINE_TABLE_SUFFIX}"
+            return SqlIdentifier(online_name, case_sensitive=True)
+        else:
+            fv_name = (
+                feature_view_name
+                if isinstance(feature_view_name, SqlIdentifier)
+                else SqlIdentifier(feature_view_name, case_sensitive=True)
+            )
+            fv_version = version if isinstance(version, FeatureViewVersion) else FeatureViewVersion(version)
+            physical_name = FeatureView._get_physical_name(fv_name, fv_version).resolved()
+            online_name = f"{physical_name}{_ONLINE_TABLE_SUFFIX}"
+            return SqlIdentifier(online_name, case_sensitive=True)
+
 
 lineage_node.DOMAIN_LINEAGE_REGISTRY["feature_view"] = FeatureView
